#!/usr/bin/env python3

import os
import json
from enum import Enum

class TimestampGranularity(Enum):
    MICROS = 1
    MILLIS = 2
    SECONDS = 3
    MINUTES = 4
    HOURS = 5
    HALF_DAYS = 6
    DAYS = 7

def get_timestamp_granularity(value):
    s = value.upper()
    ret = TimestampGranularity.MICROS
    if s == "MILLIS":
        ret = TimestampGranularity.MILLIS
    elif s == "SECONDS":
        ret = TimestampGranularity.SECONDS
    elif s == "MINUTES":
        ret = TimestampGranularity.MINUTES
    elif s == "HOURS":
        ret = TimestampGranularity.HOURS
    elif s == "HALF_DAYS":
        ret = TimestampGranularity.HALF_DAYS
    elif s == "DAYS":
        ret = TimestampGranularity.DAYS
    else:
        ret = TimestampGranularity.MICROS
    
    return ret

class configInfo:

    def __init__(self, event_reporting_wake_interval, event_reporting_minimum_count, event_reporting_flush_interval, event_reporting_trap_exceptions, event_reporting_timestamp_granularity, logging_verbose, key_caching_unstructured, key_caching_encrypt):
        self.__event_reporting_wake_interval = event_reporting_wake_interval
        self.__event_reporting_minimum_count = event_reporting_minimum_count
        self.__event_reporting_flush_interval = event_reporting_flush_interval
        self.__event_reporting_trap_exceptions = event_reporting_trap_exceptions
        self.__event_reporting_timestamp_granularity = event_reporting_timestamp_granularity
        self.__logging_verbose = logging_verbose
        self.__key_caching_unstructured = key_caching_unstructured
        self.__key_caching_encrypt = key_caching_encrypt

    def get_event_reporting_wake_interval(self):
        return self.__event_reporting_wake_interval
    event_reporting_wake_interval=property(get_event_reporting_wake_interval)

    def get_event_reporting_minimum_count(self):
        return self.__event_reporting_minimum_count
    event_reporting_minimum_count=property(get_event_reporting_minimum_count)

    def get_event_reporting_flush_interval(self):
        return self.__event_reporting_flush_interval
    event_reporting_flush_interval = property(get_event_reporting_flush_interval)

    def get_event_reporting_trap_exceptions(self):
        return self.__event_reporting_trap_exceptions
    event_reporting_trap_exceptions = property(get_event_reporting_trap_exceptions)
    
    def get_event_reporting_timestamp_granularity(self):
        return self.__event_reporting_timestamp_granularity
    event_reporting_timestamp_granularity = property(get_event_reporting_timestamp_granularity)
    
    def get_logging_verbose(self):
        return self.__logging_verbose
    logging_verbose = property(get_logging_verbose)

    def get_key_caching_unstructured(self):
        return self.__key_caching_unstructured
    key_caching_unstructured = property(get_key_caching_unstructured)
    
    def get_key_caching_encrypt(self):
        return self.__key_caching_encrypt
    key_caching_encrypt = property(get_key_caching_encrypt)

    def set(self):
        return (self.__event_reporting_wake_interval != None 
                and self.__event_reporting_minimum_count != None 
                and self.__event_reporting_flush_interval != None
                and self.__logging_verbose != None)

class ubiqConfiguration(configInfo):

    def load_config_file(self, config_file):
        try:
            with open(config_file) as json_file:
                config = json.load(json_file)
                self.set_defaults()
                if 'event_reporting' in config:
                    if 'wake_interval' in config['event_reporting']:
                        self.__event_reporting_wake_interval = config['event_reporting']['wake_interval']
                    if 'minimum_count' in config['event_reporting']:
                        self.__event_reporting_minimum_count = config['event_reporting']['minimum_count']
                    if 'flush_interval' in config['event_reporting']:
                        self.__event_reporting_flush_interval = config['event_reporting']['flush_interval']
                    if 'trap_exceptions' in config['event_reporting']:
                        self.__event_reporting_trap_exceptions = config['event_reporting']['trap_exceptions']
                    if 'timestamp_granularity' in config['event_reporting']:
                        self.__event_reporting_timestamp_granularity = get_timestamp_granularity(config['event_reporting']['timestamp_granularity'])
                if 'logging' in config:
                    if 'verbose' in config['logging']:
                        self.__logging_verbose = config['logging']['verbose']
                if 'key_caching' in config:
                    if 'unstructured' in config['key_caching']:
                        self.__key_caching_unstructured = config['key_caching']['unstructured']
                    if 'encrypt' in config['key_caching']:
                        self.__key_caching_encrypt = config['key_caching']['encrypt']
        except FileNotFoundError:
            # If file doesn't exist, use defaults
            self.set_defaults()

    def set_defaults(self):
        self.__event_reporting_wake_interval = 10
        self.__event_reporting_minimum_count = 50
        self.__event_reporting_flush_interval = 90
        self.__event_reporting_trap_exceptions = False
        self.__event_reporting_timestamp_granularity = TimestampGranularity.MICROS
        self.__logging_verbose = False
        self.__key_caching_unstructured = True
        self.__key_caching_encrypt = False

    def __init__(self, config_file = None):

        self.__event_reporting_wake_interval = None
        self.__event_reporting_minimum_count = None
        self.__event_reporting_flush_interval = None
        self.__event_reporting_trap_exceptions = None
        self.__event_reporting_timestamp_granularity = None
        self.__logging_verbose = None
        self.__key_caching_unstructured = None
        self.__key_caching_encrypt = None

        if (config_file == None):
            from os.path import expanduser
            home = expanduser("~")
            config_file = os.path.join(home, ".ubiq", "configuration")

        if os.path.exists(config_file):
            self.load_config_file(config_file)
        else:
            self.set_defaults()

        configInfo.__init__(self, self.__event_reporting_wake_interval , self.__event_reporting_minimum_count, self.__event_reporting_flush_interval, self.__event_reporting_trap_exceptions, self.__event_reporting_timestamp_granularity, self.__logging_verbose, self.__key_caching_unstructured, self.__key_caching_encrypt)



import math
import requests
import json
import time
from datetime import datetime, timezone
import threading
import atexit 

from .auth import http_auth
from .version import VERSION
from .configuration import TimestampGranularity

def get_key(api_key, dataset_name, dataset_group_name, billing_action, dataset_type, key_number, user_defined):
    return "api_key='%s' datasets='%s' billing_action='%s' dataset_groups='%s' key_number='%s' dataset_type='%s' user_defined='%s" % (api_key, dataset_name, billing_action, dataset_group_name, key_number, dataset_type, user_defined)

def format_timestamp(timestamp, timestampGranularity):
    dt = datetime(timestamp.year, timestamp.month, timestamp.day, 0, 0, 0, 0, tzinfo=timezone.utc)
    # Apply from least granular to most granular
    # No switch statements till Python 3.10
    if timestampGranularity.value <= TimestampGranularity.HALF_DAYS.value:
        if timestamp.hour >= 12:
            dt = dt.replace(hour=12)
    if timestampGranularity.value <= TimestampGranularity.HOURS.value:
        dt = dt.replace(hour=timestamp.hour)
    if timestampGranularity.value <= TimestampGranularity.MINUTES.value:
        dt = dt.replace(minute=timestamp.minute)
    if timestampGranularity.value <= TimestampGranularity.SECONDS.value:
        dt = dt.replace(second=timestamp.second)
    if timestampGranularity.value <= TimestampGranularity.MILLIS.value:
        dt = dt.replace(microsecond=math.floor(timestamp.microsecond/1000)*1000)
    if timestampGranularity.value <= TimestampGranularity.MICROS.value:
        dt = dt.replace(microsecond=timestamp.microsecond)
    return dt.isoformat()


class event:
    def __init__(self, api_key, dataset_name, dataset_group_name, billing_action, dataset_type, key_number, count, user_defined):
        self.api_key = api_key
        self.dataset_name = dataset_name
        self.dataset_group_name = dataset_group_name
        self.billing_action = billing_action
        self.dataset_type = dataset_type
        self.key_number = key_number
        self.count = count
        self.first_call_timestamp = datetime.now(timezone.utc)
        self.last_call_timestamp = datetime.now(timezone.utc)
        self.user_defined =  user_defined

    def increment_count(self, val):
        self.count = self.count + val
        self.last_call_timestamp = datetime.now(timezone.utc)
        return self.count

    def serialize(self, timestampGranularity):
        return {
            'datasets': self.dataset_name,
            'dataset_groups': self.dataset_group_name,
            'dataset_type': self.dataset_type,
            'api_key': self.api_key,
            'count': self.count,
            'key_number': self.key_number,
            'action': self.billing_action,
            'product': 'ubiq-python',
            'product_version': VERSION,
            'user-agent': 'ubiq-python/' + VERSION,
            'api_version': 'V3',
            'last_call_timestamp': format_timestamp(self.last_call_timestamp, timestampGranularity),
            'first_call_timestamp': format_timestamp(self.first_call_timestamp, timestampGranularity),
            'user_defined': self.user_defined
        }


class events:
    def __init__(self, creds, config):
        self.events_dict = {}
        self.lock = threading.Lock()
        self.count = 0

        self.config = config

        self._host = creds.host
        if (not self._host.lower().startswith('http')):
            self._host = "https://" + self._host

        self._papi = creds.access_key_id
        self._sapi = creds.secret_signing_key

        self.user_defined = {}

    def add_event(self, api_key, dataset_name, dataset_group_name, billing_action, dataset_type, key_number, count):
        self.lock.acquire()
        try:
            key = get_key(api_key, dataset_name, dataset_group_name,
                         billing_action, dataset_type, key_number, self.user_defined)
            current_count = self.events_dict.get(key, event(api_key, dataset_name, dataset_group_name,
                                                           billing_action, dataset_type, key_number, 0, self.user_defined))
            current_count.increment_count(count)
            self.events_dict.update({key: current_count})
            self.count += count
        except Exception as e:
            self.handle_exception(e)
        finally:
            self.lock.release()

    def list_events(self):
        return list(map(lambda e: e.serialize(self.config.get_event_reporting_timestamp_granularity()), self.events_dict.values()))
    
    def get_events_count(self):
        return self.count

    def add_user_defined_metadata(self, data):
        if type(data) != str:
            self.handle_exception(Exception('User defined Metadata must be a string.'))
        if len(data) > 1024:
            self.handle_exception(Exception('User defined Metadata cannot be longer than 1024 characters'))

        try:
            parsed_data = json.loads(data)
            self.user_defined = parsed_data
            print("Added user defined metadata: %s" %(data))
        except Exception as e:
            self.handle_exception(Exception('User defined Metadata must be a valid Json object'))
    
    def process_events(self):
        if (self.get_events_count() == 0 or len(self.events_dict) == 0):
            if self.config.get_logging_verbose():
                print('No events, skipping processing.')
            return
        
        if self.config.get_logging_verbose():
            print(f'Processing {self.count} events')
        
        self.lock.acquire()

        try:
            usage = json.dumps({'usage': self.list_events()})
            self.events_dict = {}
            self.count = 0

        except Exception as e:
            self.handle_exception(e)
        finally: 
            self.lock.release()

        try:
            requests.post(f'{self._host}/api/v3/tracking/events',
                        data=usage.encode('utf-8'),
                        auth=http_auth(self._papi, self._sapi)
                        )
            if self.config.get_logging_verbose():
                print(f'Processed events: {usage}')
        except Exception as e:
            self.handle_exception(e)
            
    def handle_exception(self, ex):
        if self.config.get_logging_verbose():
                        template = "An exception of type {0} occurred. Arguments:\n{1!r}"
                        message = template.format(type(ex).__name__, ex.args)
                        print(message)
        if self.config.get_event_reporting_trap_exceptions():
            pass
        else:
            raise
        


class eventsProcessor:
    def __init__(self, configuration, events):
        self.config = configuration
        self.events = events
        self._processThread = None
        
        self.flush_interval = self.config.get_event_reporting_flush_interval()
        self.next_flush = time.time() + self.flush_interval

        self.wake_interval = self.config.get_event_reporting_wake_interval()
        self.next_wake = time.time() + self.wake_interval

        self.running = False

    def start(self):
        self._processThread = threading.Thread(target=self.process)
        atexit.register(self.graceful_close)
        self._processThread.daemon = True
        self.running = True
        self._processThread.start()

    def process(self):
        while True:
            if not self.running:
                break
            
            # Run every wake
            if time.time() >= self.next_wake:
                # but only submit if events > min or it's flush time
                if self.events.get_events_count() >= self.config.get_event_reporting_minimum_count() or time.time() >= self.next_flush:
                    self.events.process_events()
                    self.next_flush = time.time() + self.flush_interval
                self.next_wake = time.time() + self.wake_interval
            
            time.sleep(self.wake_interval)
        return

    def graceful_close(self):
        if self.config.get_logging_verbose():
            print('Closing event processor')
        self.running = False
        self.events.process_events()
    
    def __del__(self):
        self.graceful_close()


#!/usr/bin/env python3

import atexit
import configparser
import os
from . import UBIQ_HOST
from .configuration import ubiqConfiguration
from .events import events, eventsProcessor

class credentialsInfo:

    def __init__(self, access_key_id, secret_signing_key, secret_crypto_access_key, host):
        self.__access_key_id = access_key_id
        self.__secret_signing_key = secret_signing_key
        self.__secret_crypto_access_key = secret_crypto_access_key
        self.__host = host

        # Event Tracking
        self.__configuration = ubiqConfiguration()
        self.__events = events(self, self.__configuration)
        self.__eventsProcessor = eventsProcessor(self.__configuration, self.__events)
        self.__eventsProcessor.start()

    def get_access_key_id(self):
        return self.__access_key_id
    access_key_id=property(get_access_key_id)

    def get_secret_signing_key(self):
        return self.__secret_signing_key
    secret_signing_key=property(get_secret_signing_key)

    def get_secret_crypto_access_key(self):
        return self.__secret_crypto_access_key
    secret_crypto_access_key = property(get_secret_crypto_access_key)

    def get_host(self):
        return self.__host
    host = property(get_host)

    def set(self):
        return (self.__access_key_id != None and self.__secret_signing_key != None and self.__secret_crypto_access_key != None and 
            self.__access_key_id.strip() != "" and self.__secret_signing_key.strip() != "" and self.__secret_crypto_access_key.strip() != "")
    
    def add_reporting_user_defined_metadata(self, data):
        self.__events.add_user_defined_metadata(data)
    
    def get_copy_of_usage(self):
        return self.__events.list_events()

    # Forward events to event queue
    def add_event(self, dataset_name, dataset_group_name, billing_action, dataset_type, key_number, count):
        return self.__events.add_event(self.__access_key_id, dataset_name, dataset_group_name, billing_action, dataset_type, key_number, count)

class configCredentials(credentialsInfo):

    def load_config_file(self, config_file, profile):
        config = configparser.ConfigParser()
        config.read(config_file)

        # Create empty dictionaries for the default and supplied profile
        d = {}
        p = {}

        # get the default profile if there is one
        if (config.has_section('default')):
            d = config['default']

        # get the supplied profile if there is one
        if (config.has_section(profile)):
            p = config[profile]

        # Use given profile if it is available, otherwise use default.
        self.__access_key_id= p.get('access_key_id', d.get('access_key_id'))
        self.__secret_signing_key = p.get('secret_signing_key', d.get('secret_signing_key'))
        self.__secret_crypto_access_key = p.get('secret_crypto_access_key', d.get('secret_crypto_access_key'))
        self.__host = p.get('SERVER', d.get('SERVER', UBIQ_HOST))


    def __init__(self, config_file = None, profile = "default"):

        self.__access_key_id = None
        self.__secret_signing_key = None
        self.__secret_crypto_access_key = None
        self.__host = None

        if (config_file == None):
            from os.path import expanduser
            home = expanduser("~")
            config_file = os.path.join(home, ".ubiq", "credentials")

        if os.path.exists(config_file):
            self.load_config_file(config_file, profile)

        credentialsInfo.__init__(self, self.__access_key_id , self.__secret_signing_key, self.__secret_crypto_access_key, self.__host)

        if (not self.set()):
            if (self.__access_key_id == None or self.__access_key_id.strip() == ""):
               raise RuntimeError("Unable to open credentials file '{0}' or unable to find 'access_key_id' value in profile '{1}'.".format(config_file, profile))
            elif (self.__secret_signing_key == None or self.__secret_signing_key.strip() == ""):
               raise RuntimeError("Unable to open credentials file '{0}' or unable to find 'secret_signing_key' value in profile '{1}'.".format(config_file, profile))
            elif(self.__secret_crypto_access_key == None or self.__secret_crypto_access_key.strip() == ""):
               raise RuntimeError("Unable to open credentials file '{0}' or unable to find 'secret_crypto_access_key' value in profile '{1}'.".format(config_file, profile))



class credentials(credentialsInfo):

    def __init__(self, access_key_id = None, secret_signing_key = None, secret_crypto_access_key = None, host = UBIQ_HOST):
        # If supplied value is None, use ENV variable, otherwise use supplied value.
        # If env value isn't set, use the supplied value anyways (None) but prevent an exception
        self.__access_key_id = (access_key_id, os.getenv('UBIQ_ACCESS_KEY_ID', access_key_id)) [access_key_id == None]
        self.__secret_signing_key = (secret_signing_key, os.getenv('UBIQ_SECRET_SIGNING_KEY', secret_signing_key)) [secret_signing_key == None]
        self.__secret_crypto_access_key = (secret_crypto_access_key, os.getenv('UBIQ_SECRET_CRYPTO_ACCESS_KEY', secret_crypto_access_key)) [secret_crypto_access_key == None]
        self.__host = (host, os.getenv('UBIQ_SERVER', host)) [host == None]

        credentialsInfo.__init__(self, self.__access_key_id,
                                 self.__secret_signing_key,
                                 self.__secret_crypto_access_key, 
                                 self.__host)
        
        if (not self.set()):
            if (self.__access_key_id == None or self.__access_key_id.strip() == ""):
               raise RuntimeError("Environment variable for 'UBIQ_ACCESS_KEY_ID' not set.")
            elif (self.__secret_signing_key == None or self.__secret_signing_key.strip() == ""):
               raise RuntimeError("Environment variable for 'UBIQ_SECRET_SIGNING_KEY' not set.")
            elif(self.__secret_crypto_access_key == None or self.__secret_crypto_access_key.strip() == ""):
               raise RuntimeError("Environment variable for 'UBIQ_SECRET_CRYPTO_ACCESS_KEY' not set.")



#!/usr/bin/env python3

import math
import sys
import typing

from M2Crypto import EVP

DEFAULT_ALPHABET: typing.Final[str] = '0123456789abcdefghijklmnopqrstuvwxyz'

class Context:
    def __init__(self,
                 key, twk,
                 maxtxtlen, mintwklen, maxtwklen,
                 radix, alpha):
        self.BLKSZ = (int)(16)
        self.key = key

        self.alg = None
        if len(key) == 16:
            self.alg = 'aes_128_cbc'
        elif len(key) == 24:
            self.alg = 'aes_192_cbc'
        elif len(key) == 32:
            self.alg = 'aes_256_cbc'
        if self.alg == None:
            raise RuntimeError('Key length invalid.')

        if radix < 2 or radix > len(alpha):
            raise RuntimeError('Unsupported radix or incompatible alphabet')

        self.alpha = alpha

        #
        # for both ff1 and ff3-1: radix**minlen >= 1000000
        #
        # therefore:
        #   minlen = ceil(log_radix(1000000))
        #          = ceil(log_10(1000000) / log_10(radix))
        #          = ceil(6 / log_10(radix))
        #
        mintxtlen = math.ceil(6 / math.log10(radix));
        if mintxtlen < 2 or mintxtlen > maxtxtlen:
            raise RuntimeError('Invalid text length bounds')

        if (mintwklen > maxtwklen or
            len(twk) < mintwklen or
            (maxtwklen > 0 and len(twk) > maxtwklen)):
            raise RuntimeError('Invalid tweak length or bounds')

        self.radix = radix

        self.mintxtlen = mintxtlen
        self.maxtxtlen = maxtxtlen
        self.mintwklen = mintwklen
        self.maxtwklen = maxtwklen

        self.twk = twk

    def PRF(self, buf):
        BLKSZ = self.BLKSZ

        if len(buf) % BLKSZ != 0:
            raise RuntimeError(
                'Plaintext length must be a multiple of ' +
                str(BLKSZ))

        cipher = EVP.Cipher(alg=self.alg, key=self.key, iv=bytearray(16), op=1)
        dst = cipher.update(buf)  

        return dst[-BLKSZ:]

    def Ciph(self, buf):
        return self.PRF(buf[0:self.BLKSZ])

def StringToNumber(radix, alpha, s):
    p = 1
    n = 0
    for i in range(len(s)):
        x = alpha.index(s[len(s) - i - 1])
        n += x * p
        p *= radix
    return n

def NumberToString(radix, alpha, n, l = 1):
    s = ''
    while n:
        s = alpha[int(n % radix)] + s
        n //= radix
    while len(s) < l:
        s = alpha[0] + s
    return s



#!/usr/bin/env python3

import unittest

import importlib
ffx = importlib.import_module('ubiq_security.structured.lib.ffx')

class TestFFX(unittest.TestCase):
    def test_context(self):
        self.assertIsNotNone(
            ffx.Context(bytes([0]*32), bytes([0]*7),
                        2**32,
                        0, 7,
                        10, ffx.DEFAULT_ALPHABET))



#!/usr/bin/env python3

import math

from . import ffx

class Context:
    def __init__(self,
                 key, twk,
                 mintwklen, maxtwklen,
                 radix, alpha = ffx.DEFAULT_ALPHABET):
        self.ffx = ffx.Context(key, twk,
                               2**32,
                               mintwklen, maxtwklen,
                               radix, alpha)

    def cipher(self, X, T, ENC):
        BLKSZ = self.ffx.BLKSZ

        n = len(X)
        u = int(n / 2)
        v = n - u

        b = int((math.ceil(math.log2(self.ffx.radix) * v) + 7) / 8)
        d = 4 * int((b + 3) / 4) + 4

        R = bytearray([0] * int((d + (BLKSZ - 1)) / BLKSZ) * BLKSZ)

        if T == None:
            T = self.ffx.twk
        if T == None:
            T = bytes([])

        if (n < self.ffx.mintxtlen or
            n > self.ffx.maxtxtlen or
            len(T) < self.ffx.mintwklen or
            (self.ffx.maxtwklen > 0 and
             len(T) > self.ffx.maxtwklen)):
            raise RuntimeError('Input or tweak length error')

        PQ = bytearray(
            [0] * (BLKSZ + int((len(T) + b + 1 + 15) / BLKSZ) * BLKSZ))

        # initialize the P portion of PQ
        PQ[:8] = [1, 2, 1,
                  self.ffx.radix >> 16 & 0xff,
                  self.ffx.radix >> 8 & 0xff,
                  self.ffx.radix & 0xff,
                  10, u & 0xff]
        PQ[8:12] = n.to_bytes(4, byteorder='big')
        PQ[12:16] = len(T).to_bytes(4, byteorder='big')

        # initialize the constant portion of Q
        PQ[BLKSZ:BLKSZ + len(T)] = T

        nA = ffx.StringToNumber(self.ffx.radix, self.ffx.alpha, X[:u])
        nB = ffx.StringToNumber(self.ffx.radix, self.ffx.alpha, X[u:])
        if not ENC:
            nA, nB = nB, nA

        mU = self.ffx.radix ** u
        mV = mU
        if u != v:
            mV *= self.ffx.radix

        for i in range(10):
            if ENC:
                PQ[-b - 1] = i
            else:
                PQ[-b - 1] = 9 - i

            PQ[-b:] = nB.to_bytes(b, byteorder='big')

            R[:BLKSZ] = self.ffx.PRF(PQ)

            for j in range(int(len(R) / BLKSZ) - 1):
                w = int.from_bytes(R[12:BLKSZ], byteorder='big')

                R[12:BLKSZ] = (w ^ (j + 1)).to_bytes(4, byteorder='big')
                R[BLKSZ * (j + 1):BLKSZ * (j + 2)] = self.ffx.Ciph(R)
                R[12:BLKSZ] = w.to_bytes(4, byteorder='big')

            y = int.from_bytes(R[:d], byteorder='big')
            if ENC:
                y = nA + y
            else:
                y = nA - y

            nA, nB = nB, nA

            if int(ENC) == i % 2:
                nB = y % mV
            else:
                nB = y % mU

        if not ENC:
            nA, nB = nB, nA

        return (ffx.NumberToString(self.ffx.radix, self.ffx.alpha, nA, u) +
                ffx.NumberToString(self.ffx.radix, self.ffx.alpha, nB, v))

    def Encrypt(self, pt, twk = None):
        return self.cipher(pt, twk, True)

    def Decrypt(self, ct, twk = None):
        return self.cipher(ct, twk, False)



#/usr/bin/env python3

import base64

from ..credentials import credentials


from .common import fmtInput, strConvertRadix, decKeyNumber, fmtOutput
from .common import fetchDataset, fetchKey
from .lib import ff1

class Decryption:
    def __del__(self):
        # todo: overwrite 'unwrapped_data_key'
        return

    def __init__(self, creds, dataset_name):
        if not creds.set():
            raise RuntimeError("credentials not set")
        
        self._creds = creds

        # If the host does not begin with either http or https
        # insert https://
        self._host = creds.host
        if (not self._host.lower().startswith('http')):
            self._host = "https://" + self._host
        self._host += '/api/v0/'

        self._papi = creds.access_key_id
        self._sapi = creds.secret_signing_key
        self._srsa = creds.secret_crypto_access_key

        self._dataset = fetchDataset(self._host, self._papi, self._sapi, dataset_name)

    def Cipher(self, ct, twk = None):
        pth = self._dataset['passthrough']
        ics = self._dataset['input_character_set']
        ocs = self._dataset['output_character_set']
        rules = self._dataset.get('passthrough_rules', [])

        input_min = self._dataset['min_input_length']
        input_max = self._dataset['max_input_length']
        
        fmt, ct, rules = fmtInput(ct, pth, ocs, ics, rules)

        input_len = len(ct)
        if input_len < input_min or input_len > input_max:
            raise RuntimeError('Invalid input len (%s) min: %s max %s'%(input_len, input_min, input_max))

        ct, n = decKeyNumber(ct, ocs, self._dataset['msb_encoding_bits'])
        if not hasattr(self, '_key') or self._key['key_number'] != n:
            self._key = fetchKey(self._host,
                                 self._papi, self._sapi, self._srsa,
                                 self._dataset['name'], n)
            if self._dataset['encryption_algorithm'] == 'FF1':
                self._ctx = ff1.Context(
                    self._key['unwrapped_data_key'],
                    base64.b64decode(self._dataset['tweak']),
                    self._dataset['tweak_min_len'], self._dataset['tweak_max_len'],
                    len(ics), ics)
            else:
                raise RuntimeError('unsupported algorithm: ' +
                                   self._dataset['encryption_algorithm'])
        ct = strConvertRadix(ct, ocs, ics)

        pt = self._ctx.Decrypt(ct, twk)

        self._creds.add_event(dataset_name=self._dataset['name'], dataset_group_name="", billing_action="decrypt",
            dataset_type="structured", key_number=n, count=1)
        return fmtOutput(fmt, pt, pth, rules)

def Decrypt(creds, dataset_name, ct, twk = None):
    return Decryption(creds, dataset_name).Cipher(ct, twk)



#/usr/bin/env python3

import base64
from .lib import ff1

from ..credentials import credentials

from .common import fmtInput, strConvertRadix, encKeyNumber, fmtOutput
from .common import fetchDataset, fetchKey, fetchCurrentKeys

class Encryption:
    def __del__(self):
        # todo: overwrite 'unwrapped_data_key'
        return

    def __init__(self, creds, dataset_name):
        if not creds.set():
            raise RuntimeError("credentials not set")
        
        self._creds = creds

        # If the host does not begin with either http or https
        # insert https://
        self._host = creds.host
        if (not self._host.lower().startswith('http')):
            self._host = "https://" + self._host
        self._host += '/api/v0/'

        self._papi = creds.access_key_id
        self._sapi = creds.secret_signing_key
        self._srsa = creds.secret_crypto_access_key

        self._dataset = fetchDataset(self._host, self._papi, self._sapi, dataset_name)
        self._key = fetchKey(self._host,
                             self._papi, self._sapi,
                             self._srsa,
                             dataset_name)

        if self._dataset['encryption_algorithm'] == 'FF1':
            self._algo = ff1.Context(
                self._key['unwrapped_data_key'],
                base64.b64decode(self._dataset['tweak']),
                self._dataset['tweak_min_len'], self._dataset['tweak_max_len'],
                len(self._dataset['input_character_set']),
                self._dataset['input_character_set'])
        else:
            raise RuntimeError('unsupported algorithm: ' +
                               self._dataset['encryption_algorithm'])

    def Cipher(self, pt, twk = None):
        pth = self._dataset['passthrough']
        ics = self._dataset['input_character_set']
        ocs = self._dataset['output_character_set']
        rules = self._dataset.get('passthrough_rules', [])

        input_min = self._dataset['min_input_length']
        input_max = self._dataset['max_input_length']

        fmt, pt, rules = fmtInput(pt, pth, ics, ocs, rules)

        input_len = len(pt)
        if input_len < input_min or input_len > input_max:
            raise RuntimeError('Invalid input len (%s) min: %s max %s'%(input_len, input_min, input_max))

        ct = self._algo.Encrypt(pt, twk)

        ct = strConvertRadix(ct, ics, ocs)
        ct = encKeyNumber(ct, ocs,
                          self._key['key_number'],
                          self._dataset['msb_encoding_bits'])
        
        self._creds.add_event(dataset_name=self._dataset['name'], dataset_group_name="", billing_action="encrypt",
                dataset_type="structured", key_number=self._key['key_number'], count=1)
        
        return fmtOutput(fmt, ct, pth, rules)
    
    def CipherForSearch(self, pt, twk=None):
        keys = fetchCurrentKeys(self._host,
                            self._papi, self._sapi,
                            self._srsa,
                            self._dataset['name'])
        
        pth = self._dataset['passthrough']
        ics = self._dataset['input_character_set']
        ocs = self._dataset['output_character_set']
        rules = self._dataset.get('passthrough_rules', [])
        

        fmt, pt, rules = fmtInput(pt, pth, ics, ocs, rules)


        searchCipher = []
        for _, (key_num, key) in enumerate(keys.items()):
            algo = ff1.Context(
                key['unwrapped_data_key'],
                base64.b64decode(self._dataset['tweak']),
                self._dataset['tweak_min_len'], self._dataset['tweak_max_len'],
                len(ics),
                ics)
            ct = algo.Encrypt(pt, twk)
            ct = strConvertRadix(ct, ics, ocs)
            ct = encKeyNumber(ct, ocs,
                          key_num,
                          self._dataset['msb_encoding_bits'])
            searchCipher.append(fmtOutput(fmt, ct, pth, rules))

        return searchCipher



def Encrypt(creds, dataset_name, pt, twk = None):
    return Encryption(creds, dataset_name).Cipher(pt, twk)

def EncryptForSearch(creds, dataset_name, pt, twk = None):
    return Encryption(creds, dataset_name).CipherForSearch(pt, twk)



import os
import setuptools

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

os.chdir(here)


with open(os.path.join(here, "README.md"), "r", encoding="utf-8") as f:
    long_description = f.read()

version_contents = {}
with open(os.path.join(here, "ubiq_security", "version.py"), "r", encoding="utf-8") as f:
    exec(f.read(), version_contents)

requirementPath = os.path.join(here, "requirements.txt")
install_requires = []
if os.path.isfile(requirementPath):
    with open(requirementPath, "r", encoding="utf-8") as f:
        install_requires = f.read().splitlines()

setuptools.setup(
    name="ubiq-security",
    version=version_contents["VERSION"],
    author="Ubiq Security, Inc.",
    author_email="support@ubiqsecurity.com",
    description="Python client library for accessing Ubiq Platform",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/ubiqsecurity/ubiq-python",
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    license="Free To Use But Restricted",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: Free To Use But Restricted",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Security",
        "Topic :: Security :: Cryptography"
    ],
    python_requires='>=3.5',
)




import base64
import http
import json
import requests
import urllib.error
from copy import copy

from .auth import http_auth
from .algorithm import algorithm
from .configuration import ubiqConfiguration

import cryptography.exceptions as crypto_exceptions
import cryptography.hazmat.primitives as crypto
import cryptography.hazmat.primitives.serialization as serialize 
from cryptography.hazmat.backends import default_backend as crypto_backend


def fetchDecryptKey(host, papi, sapi, srsa, datakey, client_id, alg):
    config = ubiqConfiguration()
    key = {}
    # If Missing or Cache is disabled
    if (not papi in fetchDecryptKey.cache or
        not datakey in fetchDecryptKey.cache[papi] or
            not config.get_key_caching_unstructured()):

        url = host + '/api/v0/decryption/key'
        response = requests.post(
            url,
            data=json.dumps(
                {
                    'encrypted_data_key': base64.b64encode(
                        datakey).decode('utf-8')
                }).encode('utf-8'),
            auth=http_auth(papi, sapi))
        if response.status_code != http.HTTPStatus.OK:
            raise urllib.error.HTTPError(
                url, response.status_code,
                http.HTTPStatus(response.status_code).phrase,
                response.headers, response.content)

        content = json.loads(response.content.decode('utf-8'))

        key['algo'] = algorithm(alg)
        # the client's id for recognizing key reuse
        key['client_id'] = client_id
        # the server's id for sending updates
        key['finger_print'] = content['key_fingerprint']
        key['session'] = content['encryption_session']

        # Need these from the response or we can't decrypt the key later
        key['encrypted_private_key'] = content['encrypted_private_key']
        key['wrapped_data_key'] = content['wrapped_data_key']

        # this key hasn't been used (yet)
        key['uses'] = 0

        if config.get_key_caching_unstructured():
            # Store in Cache
            if not papi in fetchDecryptKey.cache:
                fetchDecryptKey.cache[papi] = {}

            fetchDecryptKey.cache[papi][datakey] = key

    if len(key) == 0:
        # Get key from cache, key was not fetched.
        key = fetchDecryptKey.cache[papi][datakey]
    
    # Get a copy of the key. If it came from cache, it's a reference.
    # Decrypting the key would modify the cache, ignoring configuration.
    key = copy(key)

    if "raw" in key:
        # Unencrypted, return key
        return key
    else:
        # decrypt the client's private key (sent
        # by the server)
        prvkey = serialize.load_pem_private_key(
            key['encrypted_private_key'].encode('utf-8'),
            srsa.encode('utf-8'),
            crypto_backend())
        # use the private key to decrypt the data key
        key['raw'] = prvkey.decrypt(
            base64.b64decode(key['wrapped_data_key']),
            crypto.asymmetric.padding.OAEP(
                mgf=crypto.asymmetric.padding.MGF1(
                    algorithm=crypto.hashes.SHA1()),
                algorithm=crypto.hashes.SHA1(),
                label=None))
        
        if config.get_key_caching_unstructured() and not config.get_key_caching_encrypt():
            # Update store to have unwrapped version of the
            fetchDecryptKey.cache[papi][datakey] = key
        
        return key

fetchDecryptKey.cache = {}
